%% tclp__functors_types.pl
%% Copyright (C) 2001,2002 INRIA Emmanuel Coquery

%% This library is free software; you can redistribute it and/or
%% modify it under the terms of the GNU Lesser General Public
%% License as published by the Free Software Foundation; either
%% version 2.1 of the License, or (at your option) any later version.

%% This library is distributed in the hope that it will be useful,
%% but WITHOUT ANY WARRANTY; without even the implied warranty of
%% MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
%% Lesser General Public License for more details.

%% You should have received a copy of the GNU Lesser General Public
%% License along with this library; if not, write to the Free Software
%% Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%                                                                           %%
%%                                  TCLP                                     %%
%%                                                                           %%
%%         Emmanuel Coquery, projet Contraintes, INRIA Rocquencourt          %%
%%                                                                           %%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%% $Header: /home/beaune/contraintes/coquery/cvs/contraintes/tclpchr/src/tclp/tclp__functor_types.pl,v 1.16 2004/02/06 16:57:08 coquery Exp $

:- module(tclp__functor_types,[
                               apply/4, applyG/4,
                               % reduce_all/1, reduce_all/0,
                               multi_functor_reinit/0,
                               % functor_labelling/0,
                               try_reduce/0,
                               set_label_functors/1,
                               print_functor_types/1,
                               use_heuristic/1,
                               set_look_ahead/1
                              ]).

:- use_module(library(chr)).
:- use_module(library(lists)).
:- use_module(tclp(tclp__def)).
:- use_module(tclp(tclp__clp_types)).
:- use_module(tclp(tclp__reader)).
:- use_module(tclp(tclp__declarations)).
:- use_module(tclp(tclp__errors)).
:- use_module(tclp(tclp__printer)).
:- use_module(tclp(tclp__utils)).
:- use_module(tclp(tclp__typing_environment)).

%% Verbose modes : debug_functor_types defaults

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%% error handling
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

:- dynamic error_on_no_type/0.

%% changes error_on_no_type
set_error_on_no_type(DoIt) :-
        (
          DoIt
        ->
          asserta(error_on_no_type)
        ;
          retractall(error_on_no_type)
        ).

%% prints error with an exception or fails silently
no_type_error(Location, Message) :-
        error_on_no_type ->
        print_env(Location),
        print_error(Location, Message).

:- set_error_on_no_type(true).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%% Handler for overloading
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

handler overload.

constraints
            %% multi_type(Id, L) means that for the given type scheme
            %% AbsSc, L is the list of possible schemes
            multi_type/2,
            %% abstract_type(Id, Scheme,FSpec,Loc) associate a variable that
            %% stands for the abstract scheme
            abstract_type/4,
            %% enable_redude enables the preparation of reduce checking
            enable_reduce/0,
            %% do_reduce(X,Y), means that one can reduce more, X being the
            %% number of already done reductions and Y being the total number
            %% of reductions
            do_reduce/2,
            %% A potentialy reductible multi_type constraint
            reductible_multi_type/2,
            %% This constraint is used to 'label' one functor
            label/1,
            %% This constraint is used to 'label' the first functor in
            %% the queue
            label/0,
            %% This constraints indicate the order in which abstract schemes
            %% should be processed
            next_scheme/2,
            %% constraint used to add a scheme in the queue
            add_scheme_id/1,
            %% constraint used to remove a scheme from the queue
            delete_id/1
            .
%% special Ids first and last are used to define the begining and the end of
%% the schemes queue

%% Ensure that any abstract scheme has properly initialized variables
%init @ abstract_type(Id, VarList,FSpec ,_) ==>
%        verbose(debug_functor_types, 
%                ['freshing ',FSpec,': ',VarList]),
%        fresh_term(VarList),
%        add_scheme_id(Id).

%% rules to handle the scheme queue
addscheme @ add_scheme_id(Id) , next_scheme(Id2,last) <=>
        next_scheme(Id2,Id), next_scheme(Id,last).
remscheme @ delete_id(Id) , next_scheme(Id2,Id), next_scheme(Id,Id3) <=>
        next_scheme(Id2,Id3).

%% This rules states that a failure occurs whenever an abstract scheme has no
%% more concrete scheme.
failure @ abstract_type(Id, _, FuncDesc, Location),
          multi_type(Id,[])
          <=> %% no need to delete Id since the rule application fails
            no_type_error(Location,
                          [ 'Can''t find a good type for ', FuncDesc]),
            fail.

%% If there is only one concrete type scheme that corresponds to a abstract
%% type scheme then we instanciate it
instanciate @ multi_type(Id,[ ConcreteScheme ]),
              abstract_type(Id, AbstractScheme, FuncDesc, Location)
              <=>
                %% The Id is deleted, since we found the good type
                delete_id(Id),
                apply_scheme(ConcreteScheme, AbstractScheme,
                             FuncDesc, Location, true).

%% preparation of reduction
prepare_reduce @ enable_reduce \ multi_type(X,Y) <=>
         reductible_multi_type(X,Y).

%% finish the preparation
end_prepare @ enable_reduce <=> true.

%% procede to a reduction
reduce @ abstract_type(Id, AbstractScheme, _, _) \
         reductible_multi_type(Id,ConcreteSchemes), do_reduce(N,Total)
         <=>
           filter_schemes(AbstractScheme, ConcreteSchemes, NewConcreteSchemes),
           length(ConcreteSchemes, OldLen),
           length(NewConcreteSchemes, NewLen),
           multi_type(Id, NewConcreteSchemes),
           (
             OldLen =\= NewLen
           ->
             N1 is N+1,
             do_reduce(N1,Total)
           ;
             do_reduce(N,Total)
           ).

%% end of reductions
end_reduce @ do_reduce(N,Total) <=> N=Total.

%% to recall the last label that failed.
:- dynamic last_failed/2.
set_last_failed(Location, FuncDesc) :-
        retractall(last_failed(_,_)),
        assert(last_failed(Location, FuncDesc)).

%% labeling of functors
labeling_r @ next_scheme(first,Id) \ label <=> label(Id).
labeling @ abstract_type(Id, AbstractScheme, F/N, Location)
           \ label(Id), multi_type(Id, Types)
           <=>
             %% since we label this Id, it must be removed from the queue
             delete_id(Id),
             (
               member(OneType,Types), % choose a type ...
               multi_type(Id, [OneType]), % and try it !
               reduce_all(Remaining),
               (
                 Remaining =:= 0 -> true
               ;
                 label_functor
               )
             ;
               set_last_failed(Location, F/N), fail
             ).

%% filter_schemes(AbstractScheme, ConcreteSchemes, NewConcreteSchemes)
%% NewConcreteSchemes are the schemes in ConcreteSchemes that are compatible
%% with the constraints of the variables in AbstractScheme.
filter_schemes(_, [], []).
filter_schemes(AbstractScheme, [ConcSc|ConcreteSchemes],
               [ConcSc| NewConcreteSchemes]) :-
        check_constraint(tclp__functor_types:apply_scheme(ConcSc,
                                                          AbstractScheme,
                                                          _,_,false)),!,
        verbose(debug_functor_types,['scheme checked ',ConcSc]),
        filter_schemes(AbstractScheme, ConcreteSchemes, NewConcreteSchemes).
filter_schemes(AbstractScheme, [ConcSc|ConcreteSchemes],
               NewConcreteSchemes) :- !,
        verbose(debug_functor_types,['scheme failed ',ConcSc]),
        filter_schemes(AbstractScheme, ConcreteSchemes, NewConcreteSchemes).

%% apply_scheme([ConcreteResult | ConcreteArgs],
%%              [ AbstractResult | AbstractArgs ])
%% applies the given concrete type scheme to the given abstract type scheme

%% we consider also a special scheme for equality, named 'equality' instead
%% of a list of types

apply_scheme(equality, [ AbsResult, AbsArg1, AbsArg2 ],
             '='/2, Location, PrintError)
:-
        option(strong_eq),
        once(builtin_type(predicate,ConcResult)),
        (
          verbose(debug_functor_types,[ConcResult,' < ',AbsResult]),
          ConcResult :< AbsResult,
          verbose(debug_functor_types,ok)
        ->
          true
        ;
          (PrintError ->
              verbose(debug_functor_types,'converting for error ...'),
              term_type_down(ConcResult, ConcreteTerm),
              verbose(debug_functor_types,'concrete ok'),
              term_type_up(AbsResult, AbstractTerm),
              verbose(debug_functor_types,'abstract ok'),
              print_env(Location),
              print_error(Location, [ 'Incompatible type : ',
                                      '='/2,
                                      ' has type ',
                                      ConcreteTerm,
                                      ' but is required to have type ',
                                      AbstractTerm
                                     ]),
              fail
          )
        ),
        verbose(debug_functor_types,'result ok'),
        (
          AbsArg1 :< AbsArg2, AbsArg2 :< AbsArg1
        ->
          true
        ;
          (PrintError ->
              verbose(debug_functor_types,'converting for error ...'),
              term_type_down(AbsArg1, Down1),
              verbose(debug_functor_types,'down1 ok'),
              term_type_up(AbsArg1, Up1),
              verbose(debug_functor_types,'up1 ok'),
              verbose(debug_functor_types,'converting for error ...'),
              term_type_down(AbsArg2, Down2),
              verbose(debug_functor_types,'down2 ok'),
              term_type_up(AbsArg2, Up2),
              verbose(debug_functor_types,'up2 ok'),
              print_env(Location),
              print_error(Location, [ 'Incompatible types : the two arguments of equality must have compatible types: ',
                                      'the first argument has a type between ',
                                      Down1, ' and ', Up1,
                                      ' and the second has a type between ',
                                      Down2, ' and ', Up2 ]),
              fail
          )
        ),
        verbose(debug_functor_types,'scheme args ok').
        

%% this clause is the normal case
apply_scheme([ConcResult | ConcArgs], [ AbsResult | AbsArgs ],
             Functor/Nargs, Location, PrintError) :-
        (
          verbose(debug_functor_types,[ConcResult,' = ',AbsResult,' for ',Functor/Nargs]),
	  %% ConcResult :< AbsResult,
				% now we unify them ...
	  ConcResult = AbsResult,
          verbose(debug_functor_types,ok)
        ->
          true
        ;
          (PrintError ->
              verbose(debug_functor_types,'converting for error ...'),
              term_type_down(ConcResult, ConcreteTerm),
              verbose(debug_functor_types,'concrete ok'),
              term_type_up(AbsResult, AbstractTerm),
              verbose(debug_functor_types,'abstract ok'),
              no_type_error(Location, [ 'Incompatible type : ',
                                      Functor/Nargs,
                                      ' has type ',
                                      ConcreteTerm,
                                      ' but is required to have type ',
                                      AbstractTerm
                                     ]),
              fail
          )
        ),
        verbose(debug_functor_types,'result ok'),
        (
          PrintError
        ->
          args_data(Location, ArgsLocation, Nargs)
        ;
          true
        ),
        verbose(debug_functor_types,'args data ok'),
        apply_scheme_args(ConcArgs, AbsArgs, Functor/Nargs,
                          ArgsLocation, PrintError),
        verbose(debug_functor_types,'scheme args ok').

%% apply_scheme_args(CArgs, AArgs)
%% adds the constraint that all AArgs are :< to their corresponding type in
%% CArgs.
apply_scheme_args([],[],_,[],_).
apply_scheme_args([ ConcArg | CArgs ], [ AbsArg | AArgs ],
                  Functor/Nargs, [ Loc | Locs ], PrintError) :-
        (
          AbsArg :< ConcArg
        ->
          true
        ;
          (PrintError ->
              verbose(debug_functor_types,'converting for arg error ...'),
              term_type_up(ConcArg, ConcreteTerm),
              verbose(debug_functor_types,'concrete ok'),
              term_type_down(AbsArg, AbstractTermD),
              term_type_up(AbsArg, AbstractTermU),
              verbose(debug_functor_types,'abstract ok'),
              length(CArgs, Len),
              ArgNum is Nargs - Len,
              verbose(debug_functor_types,'argument n computed'),
              print_env(Loc),
              print_error(Loc, [ 'Incompatible type in ',
                                 Functor/Nargs,
                                 ' argument n ',
                                 ArgNum,
                                 ' has type ',
                                 AbstractTermD-AbstractTermU,
                                 ' but is required to have type ',
                                 ConcreteTerm
                               ]),
              fail
          )
        ),!,
        apply_scheme_args(CArgs,AArgs, Functor/Nargs, Locs, PrintError).

%% apply(FunctorSpec, ArgsTypes, ResultType, Location)
%% adds the constraints corresponding to the application of the functor
%% described by FunctorSpec (e.g. '+'/2), to the given type vars ArgsTypes
%% and ResultType. If there is multiple possibilities, no type inequalities
%% are really added. Reduction of possibilities may cause type constraints
%% to be added in case there is only one type left.
%% Location is a term_data that is used for reporting errors.
apply(FunctorSpec, ArgsTypes, ResultType, Location) :-
%        verbose(debug_functor_types,['starting to apply ',FunctorSpec,' ...']),
        functor_schemes(FunctorSpec, Schemes2,Location),
        (
          option(strong_eq),
          FunctorSpec = '='/2
        ->
          builtin_type(predicate,Pred),
          select_all([Pred,A,A],Schemes2,Schemes3),
          Schemes = [ equality | Schemes3 ]
        ;
          Schemes = Schemes2
        ),
%       (
%          Schemes = [ _ | _ ]
%        ->
%          verbose(debug_functor_types,
%                  [ 'Functor ',FunctorSpec,' has schemes ',
%                    Schemes])
%        ;
%          verbose(debug_functor_types, 'no scheme list')
%        ),
        multi_type(Id, Schemes),
        abstract_type(Id, [ ResultType | ArgsTypes ],
                      FunctorSpec, Location),!.
%                      FunctorSpec, Location),!,
%        verbose(debug_functor_types, 'done apply').

%% applyG(FunctorSpec, ArgsTypes, ResultType, Location)
%% Same as apply except that the used type is the generalized one
applyG(FunctorSpec, ArgsTypes, ResultType, Location) :-
        functor_schemesG(FunctorSpec, Schemes, Location),
        multi_type(Id, Schemes),
        abstract_type(Id, [ ResultType | ArgsTypes ],
                      FunctorSpec, Location),!.

%% reduce_once(TotalReductions)
%% Try to reduce the multiple type possibilities for all functors waiting.
%% Do this only once by functor. TotalReductions is the number of functors
%% which had their scheme possibilities reduced.
reduce_once(TotalReductions) :-
        verbose(debug_functor_types,'starting reduce one ...'),
        enable_reduce, do_reduce(0,TotalReductions),!,
        verbose(debug_functor_types,'reduce_one ok').

%% remaining_multi_types(Nb)
%% Nb is the number of functors that still have undetermined types
remaining_multi_types(Nb) :-
        findall_constraints(multi_type(_,_), List),
        length(List, Nb),!.
remaining_multi_types(Nb) :-
        print_error([remaining_multi_types(Nb),' failed']).

%% predicate to choose the look ahead principle
:- dynamic what_look_ahead/1.

what_look_ahead(max).
%what_look_ahead(once).
%what_look_ahead(none). %% seems faster

set_look_ahead(X) :-
        retractall(what_look_ahead(_)),
        assert(what_look_ahead(X)).

%% reduce_all(RemainingMultiTypes)
%% makes the maximum reductions of scheme possiblities.
%% RemainingMultiTypes is the number of functors that still have
%% undeterminated type schemes.
reduce_all(RemainingMultiTypes) :-
        what_look_ahead(LA),!,
        (
          LA = max ->
          (
            reduce_once(TotalReductions),
            (
              TotalReductions =:= 0
            ->
              remaining_multi_types(RemainingMultiTypes)
            ;
              reduce_all(RemainingMultiTypes)
            )
          )
        ;
          LA = once ->
          (
            reduce_once(TotalReductions) -> 
            remaining_multi_types(RemainingMultiTypes)
          )
        ;
          LA = none ->
          remaining_multi_types(RemainingMultiTypes)
        ;
          print_error(['Unknow look ahead : ',LA])
        ).

%% print_multi_types prints the functors with their line numbers and
%% remaining types

print_types([]).
print_types([ [Result|Args] | Rest ]) :-
        write('  '),
        (
          Args == []
        ;
          write(Args), write(' -> ')
        ),
        write(Result),nl,
        print_types(Rest).

print_multi_types([]).
print_multi_types([ abstract_type(Id, _, F/N, TermData) # _ | Rest ]) :-
        chr:find_constraint(Id,multi_type(Id,Types),_),
        line_number(TermData, Line),
        write(F/N), write(' (line '), write(Line), write(') :'),nl,
        print_types(Types),nl,
        print_multi_types(Rest).

%% reduce_all tests if reduce_all/1 was successful
%% or makes an error and print the remaining functors

reduce_all :-
        reduce_all(RemainingNumber),
        (
          RemainingNumber =:= 0 -> true
        ;
          chr:findall_constraints(abstract_type(_,_,_,_),AbstractConstraints),
          write('The following functors have multiple type definitions: '),nl,
          print_multi_types(AbstractConstraints),
          print_error('multiple defintions')
        ).

%% tries to find a correct typing for the predicates that have multiple types
functor_labelling :-
        verbose(debug_functor_types,'starting labelling ...'),
        set_error_on_no_type(false),
        label_functor,
        set_error_on_no_type(true),
        verbose(debug_functor_types,'labelling ok')
        ;
        set_error_on_no_type(true),
        last_failed(Loc,Desc),
%        tclp__clp_types:print_constraints,
        print_env(Loc),
        print_error(Loc, [ 'Can''t find a good type for ', Desc]),
        fail.

%% dynamic predicate on the use of the heuritic
:- dynamic use_heuristic/0.

use_heuristic(true) :- assert(use_heuristic).
use_heuristic(false) :- retractall(use_heuristic).

%% choose a functor and tries to instanciate it
label_functor :-
        (use_heuristic
        ->
        %% new version : use the queue
        label
        ;
        %% find a type to reduce
        chr:find_constraint(abstract_type(Id, _, _, _),_),!,
        label(Id)),!.
        

%% make all possible reductions and then either try to label functors or
%% prints an error in case there are still multiple types remaining
:- dynamic do_label_functors/0.

set_label_functors(DoIt) :-
        (
          DoIt
        ->
          asserta(do_label_functors)
        ;
          retractall(do_label_functors)
        ).

try_reduce :- 
        reduce_all(RemainingNumber),
        (
          RemainingNumber =:= 0
        ;
          (
            do_label_functors
          ->
            functor_labelling
          ;
            reduce_all
          )
        ),
        (option(print_stats) ->
            chr:findall_constraints(tclp__parameter(_,top,_,_,_),L),
            length(L,N),write(N),write(' top'),nl,
            chr:findall_constraints(tclp__parameter(_,_,_,_,bottom),L2),
            length(L2,N2),write(N2),write(' bottom'),nl,
            chr:findall_constraints(tclp__parameter(_,top,_,_,bottom),L3),
            length(L3,N3),write(N3),write(' both'),nl,
            chr:findall_constraints(tclp__parameter(_,_,_,_,_),L6),
            length(L6,N6),write(N6),write(' total'),nl;
            true),            
	try_bound_enumeration, !.

%% type scheme description

%% user defined types
functor_schemes(FuncDesc, Types, _Location) :-
        verbose(debug_functor_types, ['asking for ',FuncDesc]),
        user_functor_schemes(FuncDesc, Types),!,
        verbose(debug_functor_types, ['found ',FuncDesc]).

%% default types
functor_schemes(Functor, Type, Location) :-
        default_functor_schemes(Functor, Type, Location).

%% built-in schemes for int, float, byte, in_byte, character and in_character
default_functor_schemes(X/0, [[Type]],_) :-
        integer(X), !,
        (
          0 =< X, X < 256 -> builtin_type(byte,Type) ;
          X == -1 -> builtin_type(in_byte,Type) ;
          builtin_type(integer, Type)
        ).
default_functor_schemes(X/0, [[Float]],_) :-
        float(X), !, builtin_type(float, Float).
default_functor_schemes(F/0, [[Type]],_) :- !,
        (
          atom_length(F,1) -> builtin_type(character,Type) ;
          F == end_of_file -> builtin_type(in_character,Type)
        ;
          builtin_type(atom, Type),
          verbose(defaults,[ 'Defaulting to atom for ',F/0])
        ).
default_functor_schemes(F/N, [[ Term | AllTerm]], Location) :- !,
        (
          option(default)
        ->
          builtin_type(term, Term),
          uniform_list(N,Term,AllTerm),
          verbose(defaults,[ 'Defaulting to ',AllTerm,' -> ',Term,' for ',F/N])
        ;
          print_env(Location),
          print_error(Location,['No type declared for ',F/N])
        ).

user_functor_schemes(FuncDesc, Types) :-
        get_declaration(FuncDesc, Types).

%% user defined types
functor_schemesG(FuncDesc, Types,_Location) :-
        verbose(debug_functor_types, ['asking for ',FuncDesc]),
        user_functor_schemesG(FuncDesc, Types),
        verbose(debug_functor_types, ['found ',FuncDesc]),!.

%% built-in schemes and default schemes int and float
functor_schemesG(FuncDesc, Types, Location) :-
        default_functor_schemes(FuncDesc, Types, Location).

user_functor_schemesG(FuncDesc, Types) :-
        get_declarationG(FuncDesc, Types).

%% remove the remaining constraints
multi_functor_reinit :-
        use_heuristic,!,
	remove_all_constraints(next_scheme(_,_)),
	next_scheme(first,last),
        multi_functor_reinit2.
multi_functor_reinit :-
        multi_functor_reinit2.

multi_functor_reinit2 :-
        remove_all_constraints(multi_type(_,_)),
        remove_all_constraints(abstract_type(_,_,_,_)),
        remove_all_constraints(enable_reduce),
        remove_all_constraints(do_reduce(_,_)),
        remove_all_constraints(reductible_multi_type(_,_)),
        remove_all_constraints(label(_)),
        remove_all_constraints(label),
        remove_all_constraints(delete_id(_)),!.

%% prints the types of a functor
print_functor_types(F/N) :-
        (
          copy_term(N,N2), get_declaration_enum(F/N2,_)
        ->
          get_declaration_enum(F/N,Type),
          print_functor_types2(F,Type)
        ;
          (N=0 -> true ; true),
          once(default_functor_schemes(F/N,TypeList),_),
          member(Type,TypeList),
          print_functor_types2(F,Type)
        ),
        fail
        ;
        true.

print_functor_types2(F,[ Result | Args ]) :-
        instanciate_vars([ Result | Args ]),
        TypeTerm =.. [ F | Args ],
        write(':- typeof '),
        write(TypeTerm),
        write(' is '),
        write(Result),
        write('.'),
        nl,
        fail
        ;
        true.
        


