%% tclp__handle_goals.pl
%% Copyright (C) 2001,2002 INRIA Emmanuel Coquery

%% This library is free software; you can redistribute it and/or
%% modify it under the terms of the GNU Lesser General Public
%% License as published by the Free Software Foundation; either
%% version 2.1 of the License, or (at your option) any later version.

%% This library is distributed in the hope that it will be useful,
%% but WITHOUT ANY WARRANTY; without even the implied warranty of
%% MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
%% Lesser General Public License for more details.

%% You should have received a copy of the GNU Lesser General Public
%% License along with this library; if not, write to the Free Software
%% Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%                                                                           %%
%%                                  TCLP                                     %%
%%                                                                           %%
%%         Emmanuel Coquery, projet Contraintes, INRIA Rocquencourt          %%
%%                                                                           %%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%% $Header: /home/beaune/contraintes/coquery/cvs/contraintes/tclpchr/src/tclp/tclp__handle_goals.pl,v 1.9 2004/02/06 16:57:08 coquery Exp $

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%% When a term is read from the prolog source, we might need to execute it,
%% or at least to execute some goal
%% This modules handles that.
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

:- module(tclp__handle_goals,[ handle_term/2,
                               tclp__executable/1,
                               tclp__executable/2,
                               tclp__add_hook/2,
                               tclp__add_hook/3,
                               tclp__add_hook/4,
                               tclp__add_phrase_handler/3,
                               tclp__clause/4,
                               tclp__add_clause_layout/4,
                               tclp__add_clause_layout/5,
                               tclp__valid_hook/1
                               ]).

:- use_module(tclp(tclp__errors)).
:- use_module(tclp(tclp__utils)).
:- use_module(tclp(tclp__printer)).

%% Verbose modes: debug_handle_goals

%% We have a base of goals that can be executed.
%% The goal is executed at most once.
%% The predicate executable_goal is true when a goal can be executed.
%% This predicate is dynamic, so the user can add his own executable goal.
%% This is done throught add_executable_goal/2.

:- dynamic executable_goal/1.

%% Built-in executable goals
executable_goal((A,B)) :-
        executable_goal(A), executable_goal(B) ;
        warning(['The following goal was not executed: ',(A,B)]), fail.
executable_goal(op(_, _, _)).

%% Hooks are a replacement for the goal, the hook is executed instead of
%% the goal.

:- dynamic hook_goal/3.

%% hook for ','/2 : we need to replace sub goals with their own hooks
hook_goal((A,B), Location, (HookA, HookB)) :-
        args_data(Location,[ALoc,BLoc],_),
        hook_goal(A,ALoc,HookA), hook_goal(B,BLoc,HookB).

%% get_hook(Goal,Hook) Hook is the hook for goal or Hook == Goal if Goal
%% has no hooks
get_hook(Goal,Location,Hook) :- hook_goal(Goal,Location,Hook),!.
get_hook(Goal,_,Goal).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%% Add a goal to the executable goals data base
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

%% tclp__executable(Goal) adds goal to the set of executable goals
tclp__executable(Goal) :-
        asserta(executable_goal(Goal)).
%% it is itself executable
%:- tclp__executable(tclp__executable(_)).

%% tclp__executable(Goal, Condition) adds goal to the set of executable goals
%% but the goal is considered as executable only if the Condition succeeds
tclp__executable(Goal, Condition) :-
        asserta((executable_goal(Goal) :- Condition)).
%% it is itself executable
%:- tclp__executable(tclp__executable(_,_)).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%% Handle goal hooks
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

%% dynamic predicates used to check goals
:- dynamic tclp__valid_hook/1.
tclp__valid_hook(A) :- var(A),!,fail.
tclp__valid_hook((A,B)) :- 
        tclp__valid_hook(A), tclp__valid_hook(B).
tclp__valid_hook(tclp__declarations:add_val_back(_TypedTerm, _Type,
                                                 _Location)).
tclp__valid_hook(tclp__declarations:add_pred_back(_PredType, _Location)).
tclp__valid_hook(tclp__declarations:add_order_back(_A)).
tclp__valid_hook(tclp__declarations:add_type_back(_A,_L)).
tclp__valid_hook(tclp__declarations:untyped(_A)).
tclp__valid_hook(tclp__declarations:typeof(_B is _C,_TData)).
tclp__valid_hook(tclp__declarations:type(_D,_TData2)).
tclp__valid_hook(tclp__declarations:order(_E)).
tclp__valid_hook(tclp__declarations:tclp__include(_F,_Loc)).
tclp__valid_hook(tclp__handle_goals:tclp__add_clause_layout(_C1,_L1,_H1,_B1)).
tclp__valid_hook(tclp__handle_goals:tclp__add_clause_layout(_C2,_L2,
                                                            _H2,_B2,Co2)) :-
        tclp__valid_hook(Co2).
tclp__valid_hook(tclp__declarations:tclp__add_clause_op(_Op)).
tclp__valid_hook(tclp__declarations:tclp__add_clause_op(_Op2,_Type)).
tclp__valid_hook(sicstus_modules:load_module_def(_Module)).
tclp__valid_hook(user:args_location(_Location,_)).
tclp__valid_hook(get_hook(_A,_LA,_HA)).
tclp__valid_hook(consult(_Files)).
tclp__valid_hook(tclp__clause(_Head,_HeadLoc,_HeadDef,_)).
tclp__valid_hook(op(_, _, _)).
tclp__valid_hook(absolute_file_name(_File,_F)).
tclp__valid_hook(chr__clause(_Rule, _RuleLoc, _Heads, _Bodies)).
tclp__valid_hook(tclp__executable(_,C)) :-
        tclp__valid_hook(C).
tclp__valid_hook(tclp__executable(_)).
tclp__valid_hook(executable_goal(_)).
tclp__valid_hook(tclp__add_hook(_,_)).
tclp__valid_hook(tclp__add_hook(_,_,_)).
tclp__valid_hook(tclp__add_hook(_,_,_,C)) :-
        tclp__valid_hook(C).


%% tclp__add_hook(Goal,Hook) adds a hook for the given goal.
%% adds a hook for Goal and removes its previous declarations
tclp__add_hook(Goal, Hook) :-
        tclp__add_hook(Goal,_,Hook).

%% tclp__add_hook(Goal,Location,Hook) adds a hook for the given goal.
%% adds a hook for Goal and removes its previous declarations
tclp__add_hook(Goal, Location, Hook) :-
        %retractall(user:Goal),
        verbose(debug_handle_goals,['Hook : ',Goal,' -> ',Hook]),
        asserta(hook_goal(Goal, Location, Hook)).

%% tclp__add_hook(Goal,Location, Hook, Condition) adds a hook for the given
%% goal if Condition is satisfied.
%% adds a hook for Goal and removes its previous declarations.
tclp__add_hook(Goal,Location,  Hook, Condition) :-
        %retractall(user:Goal),
        verbose( debug_handle_goals,
                ['Hook : ',Goal,' -> ',Hook,' :if: ',Condition]),
        (option(check_hooks) ->
            ( tclp__valid_hook(Condition) ;
                print_error(["Illegal hook:", Condition]) )
        ; true),
        asserta((hook_goal(Goal, Location, Hook) :- Condition)).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%% Phrase handling
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

%% hooks for entire phrases

:- dynamic dyn_handle_phrase/2.

%% tclp__add_phrase_handler(Pattern, Location, Goal)  adds a phrase handler
%% that works the following way :
%% Goal is executed after Pattern was unified with the phrase and Location
%% unified with the term data corresponding to the phrase
tclp__add_phrase_handler(Pattern, Location, Goal) :-
        asserta((dyn_handle_phrase(Pattern, Location) :- Goal)).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%% Handling any term
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

%% this predicate takes a list of goals and executes them
exec_goals([]).
exec_goals([Goal-Location|Rest]) :-
        (
          executable_goal(Goal)
        ->
          get_hook(Goal, Location, Hook), 
          verbose(debug_handle_goals,['Running hook : ',Hook]),
          (
            (option(check_hooks) ->
                (tclp__valid_hook(Hook) -> true;
                    print_error(['Illegal hook: ',Hook])); true),
            Hook ; print_error('hook failed')
          )
        ;
          true
        ),
        exec_goals(Rest).

%% This predicate looks for goals, checks whenever they can be executed
%% and execute them. If the term is a declaration of an executable goal,
%% the clause/fact is asserted in order to be able to execute goals, except
%% if the goal has been hooked

handle_term(Term, Location) :-
        dyn_handle_phrase(Term, Location).

handle_term((:- Goal),Location) :- !,
         exec_goals([Goal-Location]).

handle_term(Phrase, Location) :-
        tclp__clause(Phrase, Location, [], Bodies),!,
        exec_goals(Bodies).
        
handle_term(_,_).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%% recognizing clauses
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

:- dynamic tclp__clause/4.

%tclp__clause(X,Loc,[X-Loc-pred],[]).

tclp__add_clause_layout(Phrase,Location,Heads,Bodies) :-
        asserta((tclp__clause(Phrase,Location,Heads,Bodies) :- !)).
tclp__add_clause_layout(Phrase,Location,Heads,Bodies,Condition) :-
        asserta((tclp__clause(Phrase,Location,Heads,Bodies) :- Condition,!)).


reinit :-
        retractall(executable_goal(_)),
        retractall(hook_goal(_,_,_)),
        retractall(dyn_handle_phrase(_,_)),
        retractall(tclp__clause(_,_,_,_)),
        asserta(tclp__clause(X,Loc,[X-Loc-pred],[])),
        tclp__executable(tclp__executable(_)),
        tclp__executable(tclp__executable(_,_)),
        tclp__executable(tclp__add_hook(_,_)),
        tclp__executable(tclp__add_hook(_,_,_)),
        tclp__executable(tclp__add_hook(_,_,_,_)).

:- reinit.
